#########
# Author:        Marina Gourtovaia
# Maintainer:    $Author: ajb $
# Created:       14 October 2010
# Last Modified: $Date: 2011-07-21 16:26:34 +0100 (Thu, 21 Jul 2011) $
# Id:            $Id: fastq_old2new.pm 13725 2011-07-21 15:26:34Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-37.0/lib/npg_common/extractor/fastq_old2new.pm $
#

package npg_common::extractor::fastq_old2new;

use Carp;
use English qw(-no_match_vars);
use Moose;
use Moose::Meta::Class;
use MooseX::StrictConstructor;
use File::Spec;

use npg_common::extractor::fastq qw/split_reads/;
use npg_common::roles::run::lane::file_names;

with  'npg_common::roles::run';


use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13725 $ =~ /(\d+)/mxs; $r; };
## no critic (Documentation::RequirePodAtEnd)

Readonly::Scalar our $LANE_NUM_MIN => 1;
Readonly::Scalar our $LANE_NUM_MAX => 8;
Readonly::Scalar our $EXT => q[fastq];

=head1 NAME

npg_common::extractor::fastq_old2new

=head1 VERSION

$Revision: 13725 $

=head1 SYNOPSIS



=head1 DESCRIPTION

Splits old-style fastq files infor a separate forward and reverse fastq file and
produces fastqcheck files for new fastq files.

=head1 SUBROUTINES/METHODS

=cut

=head2 lane

Array ref for lane numbers to be split.

=cut
has 'lane'    =>      (isa       => 'ArrayRef',
                       is        => 'ro',
                       required  => 0,
                       lazy_build => 1,
                       documentation => 'Array ref for lane numbers to be split.',
                      );
sub _build_lane {
    my @a = ($LANE_NUM_MIN .. $LANE_NUM_MAX);
    return \@a;
}

=head2 dir_in

Directory with input old-style fastq files.

=cut
has 'dir_in'    =>    (isa       => 'Str',
                       is        => 'ro',
                       required  => 1,
                       documentation => 'Directory with input old-style fastq files.',
                      );

=head2 dir_out

Directory to output split files to.

=cut
has 'dir_out'    =>   (isa       => 'Str',
                       is        => 'ro',
                       required  => 1,
                       documentation => 'Directory to output split files to.',
                      );

=head2 _fastq_files

Old-style fastq files found in the input directory

=cut
has '_fastq_files' => (isa        => 'HashRef',
                       is         => 'ro',
                       required   => 0,
                       lazy_build => 1,
                      );
sub _build__fastq_files {
    my $self = shift;
    my $pattern = join q[_], $self->id_run, q[s], q[*];
    $pattern .= ".$EXT";
    my @files = glob File::Spec->catfile($self->dir_in, $pattern);
    my $h = {};
    foreach my $f (@files) {
        $h->{$f} = 1;
    }
    return $h;
}

=head2 BUILD

Sanity checking for attributes

=cut
sub BUILD {
    my $self = shift;

    foreach my $lane (@{$self->lane}) {
        if ($lane < $LANE_NUM_MIN || $lane > $LANE_NUM_MAX) {
            croak qq[Invalid lane number $lane, should be between $LANE_NUM_MIN and $LANE_NUM_MAX, inclusive];
	}
    }

    if (!-d $self->dir_in) {
        croak q[Input directory ] . $self->dir_in . q[ does not exist];
    }
    if (!-d $self->dir_out) {
        croak q[Output directory ] . $self->dir_out . q[ does not exist];
    }
    if (!-r $self->dir_out) {
        croak q[Output directory ] . $self->dir_out . q[ is not writable];
    }

    if (!scalar keys %{$self->_fastq_files}) {
        croak qq[No old-style $EXT file in the input directory ] . $self->dir_in;
    }

    return;
}

sub _filename {
    my ($self, $position, $end) = @_;
    return Moose::Meta::Class->create_anon_class(
             roles => [qw/
                          npg_common::roles::run
                          npg_common::roles::run::lane
                          npg_common::roles::run::lane::tag
                          npg_common::roles::run::lane::file_names
                         /])->new_object(
	     id_run => $self->id_run, position => $position)->generate_filename($EXT, $end);
}

=head2 convert

Converts old-style fastq input files into separate forward and reverse fastq files and produces a fastqcheck file for every new fastq file.

=cut
sub convert {
    my $self = shift;

    my $id_run = $self->id_run;
    my $dir_in = $self->dir_in;
    my $dir_out = $self->dir_out;

    foreach my $lane (@{$self->lane}) {
        my $path = File::Spec->catfile($dir_in, $self->_filename($lane)->[1]);
        if (exists $self->_fastq_files->{$path}) {
            my @fq_out = ();
            my @fqc_out = ();
            foreach my $i ((1,2)) {
                my $name = File::Spec->catfile($dir_out, $self->_filename($lane, $i)->[0]);
                push @fq_out, $name;
                push @fqc_out, $name . q[check];
	    }
            split_reads($path, [], \@fq_out, \@fqc_out);
	} else {
            carp qq[$path does not exist];
	}
    }
    return;
}

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Moose::Meta::Class

=item File::Spec

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::StrictConstructor

=item File::Spec

=item npg_common::extractor::fastq

=item npg_common::roles::run::lane::file_names

=item npg_common::roles::run

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut


