#########
# Author:        gq1
# Maintainer:    $Author: ajb $
# Created:       2010 01 15
# Last Modified: $Date: 2011-07-21 16:26:34 +0100 (Thu, 21 Jul 2011) $
# Id:            $Id: SAM_Header.pm 13725 2011-07-21 15:26:34Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-37.0/lib/npg_common/sequence/SAM_Header.pm $
#

package npg_common::sequence::SAM_Header;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use File::Basename;
use File::Spec;
use XML::LibXML;
use npg_qc::autoqc::checks::check;
use npg_common::sequence::SAM_Second_Call;
use npg_common::sequence::Fastq2SAM;
use npg_common::sequence::SAM_Index_Tag;
use npg_common::sequence::BAM_MarkDuplicate;
use npg_pipeline::archive::file::SchemaInformation;
use Cwd qw(abs_path);

use npg_qc::autoqc::qc_store::options qw/$ALL $LANES $PLEXES/;
use npg_qc::autoqc::qc_store;
use npg_common::sequence::reference;
use npg_common::sequence::reference::roles::find;
use Moose::Meta::Class;
use List::Util qw(first);

with qw{ MooseX::Getopt
         npg_common::roles::run::short_info
         npg_common::roles::run::path_info
         npg_common::roles::log
       };
with qw{ npg_common::roles::run::long_info};
with qw{ npg_common::roles::run::lane};
with qw{ npg_common::roles::run::lane::map2lims };

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13725 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $SEQUENCING_PLATFORM => q{ILLUMINA};
Readonly::Scalar our $SEQUENCING_CENTER   => q{SC};

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::sequence::SAM_Header

=head1 VERSION

$LastChangedRevision: 13725 $

=head1 SYNOPSIS

=head1 DESCRIPTION

this module try to get version number of different programs to add to a sam header section

rta, bustard and gerald from config xml

run bwa and samtools

and version numbers from module Fastq2SAM and SAM_Second_Call

=head1 SUBROUTINES/METHODS

=cut

=head2 no_alignment

skip the alignment, generate bam file without alignment

=cut

has 'no_alignment' => (isa           => q{Bool},
                       is            => q{rw},
                       required      => 0,
                       documentation => q{skip alignment},
                      );

=head2 human_split

to indicate this sam file is split or not, and human part or nonhuman part

=cut

has 'human_split'  => (isa           => q{Str},
                       is            => q{rw},
                       required      => 0,
                       documentation => q{human_split type},
                      );

=head2 human_fastq_split

fastq_split module version number

=cut

has 'human_fastq_split' => (isa           => q{HashRef},
                            is            => q{rw},
                            lazy_build    => 1,
                            documentation => q{human_fastq_split version},
                           );

sub _build_human_fastq_split{
  my $self = shift;
  my $version = npg_common::Fastq_Split->VERSION;
  return {ID=>q{human_fastq_split}, PN=>q{fastq_split}, VN=>$version};
}

=head2 tag_index

tag_index if the sam file is for one tag

=cut

has 'tag_index'    => (isa           => q{Int},
                       is            => q{rw},
                       required      => 0,
                       documentation => q{tag index},
                      );

=head2 picard_dict

A path to reference picard dictionary file

=cut

has 'picard_dict'  => (isa           => 'Str',
                       is            => 'rw',
                       required      => 0,
                       lazy_build    => 1,
                       documentation => 'Reference location for picard dictionary file',
                     );

sub _build_picard_dict {
  my $self = shift;

  if( $self->human_split() && ( $self->human_split() eq q{human} ) ){

     return $self->_get_picard_dict_by_species(q{Homo_sapiens});
  } elsif ( $self->human_split() && ( $self->human_split() eq q{phix} ) ){
     return $self->_get_picard_dict_by_species(q{PhiX});
  }

  if(! $self->id_run && ! $self->position() ){
    croak 'no id_run and position given, picard reference dictionary file cannot be retrieved';
  }

  my $ref_picard = npg_common::sequence::reference->new(
             id_run => $self->id_run(),
             position  => $self->position(),
             aligner   => q{picard},
             tag_index => $self->tag_index(),
            );
  my $refs;
  eval {
    $refs = $ref_picard->refs();
    1;
  } or do {
    my $error_message = qq{Error: picard reference dictionary cannot be retrieved; $EVAL_ERROR};
    carp $error_message;
    return q{};
  };

  if (scalar @{$refs} > 1) {
    carp q[Run ] . $self->id_run . q[ lane ] . $self->position . q[ - multiple references for a pooled lane: ] . (join q[;], @{$refs});
    return q{};
  }

  if(! scalar @{$refs}){
    return q{};
  }

  my $reference_picard_dict = pop @{$refs};
  $reference_picard_dict .= q{.dict};

  if(! -e $reference_picard_dict){
    croak q[Run ] . $self->id_run . q[ lane ] . $self->position . q[ - reference picard dictionary missing; ].$reference_picard_dict;
  }
  return $reference_picard_dict;
}

sub _get_picard_dict_by_species {
   my ($self, $species) = @_;

   my $ref_obj = Moose::Meta::Class->create_anon_class( roles =>[qw/npg_common::sequence::reference::roles::find/] )
                ->new_object({species => $species, aligner=>'picard'});

   my $ref_picard;
   eval{
        $ref_picard  = $ref_obj->refs()->[0];
        $ref_picard .= q{.dict};
        1;
   } or do {
        carp $EVAL_ERROR;
   };
   return $ref_picard;
}

=head2 split_fastq_by_tag

split_fastq_by_tag module version number

=cut

has 'split_fastq_by_tag'=> (isa           => q{HashRef},
                            is            => q{rw},
                            lazy_build    => 1,
                            documentation => q{split_fastq_by tag version},
                           );

sub _build_split_fastq_by_tag{
  my $self = shift;
  my $version = npg_common::sequence::SplitFastqByTag->VERSION;
  return {ID => q{split_fastq_by_tag},
          PN => q{split_fastq_by_tag},
          VN => $version,
          DS => q{split multiplex fastq into files per tag},
         };
}


=head2 rta

get RTA software version number from config xml file

=cut

has 'rta'          => (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{rta version},
                      );

sub _build_rta{
  my $self = shift;

  my $rta = {ID => q{SCS}};

  my $intensity_path;
  eval{
     $intensity_path = $self->intensity_path();
     1;
  } or do {
     $self->log( 'Intensity path not available' );
     return $rta;
  };

  my $rta_config_xml = $intensity_path.q{/config.xml};

  if(! -e $rta_config_xml){
    carp qq{RTA config xml NOT exist: $rta_config_xml};
    return $rta;
  }

  my $dom = XML::LibXML->new->parse_file( $rta_config_xml );
  my $ele = $dom->getElementsByTagName('Software')->[0];
  my $version = $ele->getAttribute('Version');
  my $name = $ele->getAttribute('Name');

  $rta->{VN} = $version;
  $rta->{PN} = $name;

  return $rta;
}

=head2 bustard

get bustard software version number from config xml file

=cut

has 'bustard'      => (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{bustard version},
                      );

sub _build_bustard{
  my $self = shift;

  my $bustard = {ID => q{basecalling}};
  my $bustard_path;

  eval{
    $bustard_path = $self->bustard_path();
    1;
  } or do {
      $self->log('Bustard path not available');
      return $bustard;
  };

  my $bustard_config_xml = $bustard_path.q{/config.xml};

  if(! -e $bustard_config_xml){
    $self->log(qq{Bustard config xml NOT exist: $bustard_config_xml});
    return $bustard;
  }

  my $dom = XML::LibXML->new->parse_file($bustard_config_xml );
  my $ele;
  my @childnodes = $dom->getElementsByTagName('Run')->[0]->childNodes();
  foreach my $child (@childnodes){
     if($child->nodeName() eq q{Software}){
        $ele = $child;
        last;
     }
  }

  my $version = $ele->getAttribute('Version');
  my $name = $ele->getAttribute('Name');

  $bustard->{PN} = $name;
  $bustard->{VN} = $version;

  return $bustard;
}

=head2 recalibration

recalibration program and version

=cut

has 'recalibration' => (isa           => q{HashRef},
                        is            => q{rw},
                        lazy_build    => 1,
                        documentation => q{recalibration program and version},
                      );

sub _build_recalibration{
  my $self = shift;

  my $schema_info = $self->schema_info();
  my $recalibration = $schema_info->{recalibration};
  $recalibration||= {};

  return $recalibration;
}

=head2 schema_info

 schema information object to retrieve progrm list generating bam

=cut

has 'schema_info'    => (isa           => q{HashRef},
                         is            => q{rw},
                         lazy_build    => 1,
                         documentation => q{bwa program},
                        );

sub _build_schema_info {
   my $self = shift;

   my $schema_info = {};

   eval {
     my $schema = npg_pipeline::archive::file::SchemaInformation->new(id_run        => $self->id_run(),
                                                                      lane          => $self->position(),
                                                                      archive_path  => $self->archive_path(),
                                                                     );
     my $schema_info_arrayref = $schema->read_in_from_file();

     foreach my $program ( @{$schema_info_arrayref} ) {
        delete $program->{PP};
        $schema_info->{ $program->{ID} } = $program;
     }
     1;
   } or do {
     carp $EVAL_ERROR;
   };

   return $schema_info;
}

=head2 bwa_cmd

bwa command

=cut

has 'bwa_cmd'      => (isa           => q{Str},
                       is            => q{rw},
                       documentation => q{bwa program},
                      );
=head2 bwa_aln_cmd

bwa aln command

=cut

has 'bwa_aln_cmd'  => (isa           => q{Str},
                       is            => q{rw},
                       documentation => q{bwa aln command},
                      );
=head2 bwa_sam_cmd

bwa sam pe or se command

=cut

has 'bwa_sam_cmd'  => (isa           => q{Str},
                       is            => q{rw},
                       documentation => q{bwa sam pe or se command},
                      );
=head2 bwa_aln

bwa aln header hashref

=cut
has 'bwa_aln'      => (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{bwa aln version},
                      );
sub _build_bwa_aln {
  my $self = shift;

  if(!$self->bwa_aln_cmd){
    return {};
  }

  my $bwa_version = npg_qc::autoqc::checks::check->aligner_version($self->bwa_cmd());
  my $bwa_aln = {ID => q{bwa_aln},
                 PN => q{bwa aln},
                 VN => $bwa_version,
                 CL => $self->bwa_aln_cmd(),
                 DS => q{bwa alignment from fastq files},
                };

  return $bwa_aln;
}

=head2 bwa_sam

bwa sampe or samse

=cut
has 'bwa_sam'      => (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{bwa sampe or samse version},
                      );
sub _build_bwa_sam {
  my $self = shift;

  if(!$self->bwa_sam_cmd){
    return {};
  }

  my $bwa_version = npg_qc::autoqc::checks::check->aligner_version($self->bwa_cmd());
  my $bwa_sam = {ID => q{bwa_sam},
                 PN => q{bwa sam},
                 VN => $bwa_version,
                 CL => $self->bwa_sam_cmd,
                 DS => q{bwa converting alignemtns to sam},
                };

  return $bwa_sam;
}

=head2 fastq2sam_cmd

fastq2sam command

=cut

has 'fastq2sam_cmd'  => (isa           => q{Str},
                         is            => q{rw},
                         documentation => q{fastq2sam command},
                        );
=head2 fastq2sam

fastq2sam version

=cut

has 'fastq2sam'    => (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{fastq2sam version},
                      );
sub _build_fastq2sam{
  my $self = shift;
  my $version =npg_common::sequence::Fastq2SAM->VERSION;
  return {ID => q{fastq2sam},
          PN => q{fastq2sam},
          VN => $version,
          CL => $self->fastq2sam_cmd(),
          DS => q{converts fastq to sam without any alignment},
         };
}

=head2 sam_fastq_check_cmd

sam_fastq_check command

=cut

has 'sam_fastq_check_cmd'      => (isa           => q{Str},
                                   is            => q{rw},
                                   documentation => q{sam_fastq_check command},
                                  );

=head2 sam_fastq_check

sam_fastq_check version

=cut

has 'sam_fastq_check' => (isa           => q{HashRef},
                          is            => q{rw},
                          lazy_build    => 1,
                          documentation => q{sam fastq check version},
                         );

sub _build_sam_fastq_check{
  my $self = shift;

  if(! $self->sam_fastq_check_cmd()){
    return {};
  }

  my $version = npg_common::sequence::SAM_Fastq_Check->VERSION;
  return {ID => q{sam_fastq_check},
          PN => q{sam_fastq_check},
          VN => $version,
          CL => $self->sam_fastq_check_cmd(),
          DS => q{program used to check bases and quality scores in bam are the same as the ones in original fastq files used to generate the bam},
         };
}

=head2 sam_original_quality_cmd

sam_original_quality_cmd

=cut

has 'sam_original_quality_cmd' => (isa           => q{Str},
                                   is            => q{rw},
                                   documentation => q{sam_original_quality_cmd},
                                  );

=head2 sam_original_quality

sam_original_quality version

=cut

has 'sam_original_quality' => (isa           => q{HashRef},
                               is            => q{rw},
                               lazy_build    => 1,
                               documentation => q{sam_original_quality version},
                              );

sub _build_sam_original_quality{
  my $self = shift;

  if(! $self->sam_original_quality_cmd()){
    return {};
  }

  my $version = npg_common::sequence::SAM_Original_Quality->VERSION;
  return {ID => q{sam_original_quality},
          PN => q{sam_original_quality},
          VN => $version,
          CL => $self->sam_original_quality_cmd(),
          DS => q{add original quality score for each read as OQ string if the current scores are recalibrated},
         };
}

=head2 sam_second_call_cmd

bwa command

=cut

has 'sam_second_call_cmd'      => (isa           => q{Str},
                                   is            => q{rw},
                                   documentation => q{sam_second_call command},
                                  );

=head2 sam_second_call

sam_second_call version

=cut

has 'sam_second_call' => (isa           => q{HashRef},
                          is            => q{rw},
                          lazy_build    => 1,
                          documentation => q{sam_secon_call version},
                         );

sub _build_sam_second_call{
  my $self = shift;

  if(! $self->sam_second_call_cmd()){
    return {};
  }

  my $version = npg_common::sequence::SAM_Second_Call->VERSION;
  return {ID => q{sam_second_call},
          PN=>q{sam_second_call},
          VN=>$version,
          CL =>$self->sam_second_call_cmd(),
          DS => q{add second call bases to each read as E2 string},
         };
}

=head2 sam_header

sam_header version

=cut

has 'sam_header'      => (isa           => q{HashRef},
                          is            => q{rw},
                          lazy_build    => 1,
                          documentation => q{sam header version},
                         );

sub _build_sam_header{
  my $self = shift;

  return {ID => q{sam_header},
          PN => q{sam_header},
          VN => $VERSION,
          DS => q{add PG, RG and SQ tag in bam header, and add RG tag for each read},
         };
}

=head2 samtools_fixmate

samtool fixmate program

=cut

has 'samtools_fixmate'=> (isa           => q{HashRef},
                          is            => q{rw},
                          lazy_build    => 1,
                          documentation => q{samtools fixmate program},
                         );

sub _build_samtools_fixmate{
  my $self = shift;

  if(!$self->samtools_cmd){
    return {};
  }

  my $samtools_version = npg_qc::autoqc::checks::check->aligner_version($self->samtools_cmd());
  my $samtools = {ID => q{samtools_fixmate},
                  PN => q{samtools},
                  VN => $samtools_version,
                  CL => $self->samtools_cmd.q{ fixmate - -},
                  DS => q{Samtools used to fix mate information},
                 };

  return $samtools;
}


=head2 samtools_cmd

samtools command

=cut

has 'samtools_cmd' => (isa           => q{Str},
                       is            => q{rw},
                       documentation => q{samtools program},
                      );
=head2 samtools_sort

samtool version

=cut

has 'samtools_sort'=> (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{samtools sort version},
                      );

sub _build_samtools_sort{
  my $self = shift;

  if(!$self->samtools_cmd){
    return {};
  }

  my $samtools_version = npg_qc::autoqc::checks::check->aligner_version($self->samtools_cmd());
  my $samtools = {ID => q{samtools_sort},
                  PN => q{samtools},
                  VN => $samtools_version,
                  CL => $self->samtools_cmd.q{ sort - },
                  DS => q{Samtools used to sort reads in bam},
                 };

  return $samtools;
}

=head2 picard_converter_cmd

picard_converter_cmd

=cut

has 'picard_converter_cmd' => (isa           => q{Str},
                               is            => q{rw},
                               documentation => q{picard_converter_cmd},
                              );
=head2 picard

picard version

=cut

has 'picard'       => (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{picard version},
                      );

sub _build_picard{
  my $self = shift;

  if(!$self->picard_converter_cmd){
    return {};
  }

  my $picard_version = $self->_get_picard_version_from_cmd($self->picard_converter_cmd());

  my $picard = {ID=>q{picard_samformatconverter},
                PN => q{Picard SamFormatConverter},
                VN => $picard_version,
                CL => $self->picard_converter_cmd(),
                DS => q{Picard used to convert sam to bam},
               };

  return $picard;
}


=head2 picard_markduplicates

picard markduplicates

=cut

has 'picard_markduplicates'       => (isa          => q{HashRef},
                                     is            => q{rw},
                                     lazy_build    => 1,
                                     documentation => q{picard mark duplicates version},
                                    );

sub _build_picard_markduplicates {
  my $self = shift;

  my $bam_markduplicate = npg_common::sequence::BAM_MarkDuplicate->new({
       input_bam     => q{},
       output_bam    => q{},
       metrics_file  => q{},
       temp_dir      => q{},
       metrics_json  => q{},
  });

  my $mark_duplicate_cmd = $bam_markduplicate->mark_duplicate_cmd();

  my $picard_version = $self->_get_picard_version_from_cmd($mark_duplicate_cmd);

  my $picard = {ID=>q{picard_markduplicates},
                PN => q{Picard MarkDuplicates},
                VN => $picard_version,
                CL => $mark_duplicate_cmd,
                DS => q{Picard used to mark duplicates reads in bam file},
                };

  return $picard;
}

sub _get_picard_version_from_cmd {
   my ($self, $cmd) = @_;

  my ($picard_path) = grep {/picard/mxs} split /\s/mxs, $cmd;
  $picard_path = abs_path $picard_path;
  my ($picard_version) = $picard_path =~ m/ picard-tools- ( \d+ [.] \d+ )/mxs;

  return $picard_version;
}

=head2 sam_index_tag_cmd

sam_index_tag command

=cut

has 'sam_index_tag_cmd'    => (isa           => q{Str},
                               is            => q{rw},
                               documentation => q{sam_index_tag command},
                              );
=head2 sam_index_tag

sam_index_tag script version

=cut

has 'sam_index_tag'=> (isa           => q{HashRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{sam_index_tag script version},
                      );

sub _build_sam_index_tag {
  my $self = shift;

  if( !$self->sam_index_tag_cmd() ){
    return {};
  }
  my $version =npg_common::sequence::SAM_Index_Tag->VERSION;
  return {ID => q{sam_index_tag},
          PN => q{sam_index_tag},
          VN => $version,
          CL => $self->sam_index_tag_cmd(),
          DS => q{add indexing tag sequences and quality scores to all the first reads as RT and QT string},
         };

}


=head2 sam_file

input sam file

=cut

has 'sam_file'        => (isa             => 'Str',
                          is              => 'rw',
                          documentation   => 'Input sam file, will read from stadnd input if not given',
                         );

=head2 qc_store

qc_store object

=cut

has 'qc_store'           => (isa           => 'npg_qc::autoqc::qc_store',
                             is            => 'rw',
                             lazy_build    => 1,
                             documentation => 'autoqc qc_store object',
                            );

sub _build_qc_store {
   my $self = shift;
   return npg_qc::autoqc::qc_store->new(use_db => q[no]);
}


=head2 median_insert_size

median insert_size from autoqc

=cut

has 'median_insert_size' => (isa           => 'Int',
                             is            => 'rw',
                             lazy_build    => 1,
                             documentation => 'median insert size from autoqc',
                            );

sub _build_median_insert_size {
   my $self = shift;

   my $archive_path;
   eval{
      $archive_path = $self->archive_path();
      1;
   } or do {
      carp $EVAL_ERROR;
   };

   my $qc_store =  $self->qc_store();
   my $collection;

   if( $archive_path ){
      my $qc_path_to_load = $archive_path;
      if( defined $self->tag_index() ){
         $qc_path_to_load .= q{/lane}.$self->position();
      }
      $qc_path_to_load .= q{/qc};
      $collection = $qc_store->load_from_path( $qc_path_to_load );
   }
   #else{
      # $ALL $LANES $PLEXES
      #my $type = defined $self->tag_index() ? $PLEXES
      #      :                              $LANES
      #      ;
      #$collection = $qc_store->load_run($self->id_run(), undef, [$self->position], $type);
   #}

   if(! $collection ){
     return 0;
   }
   my $collection_lane = $collection->slice(q[position], $self->position());
   my $insert_size_collection = $collection_lane->slice('check_name', 'insert size');

   my $insert_size_objs = $insert_size_collection->results();

   if( defined $self->tag_index() ){
      my $tag_index = $self->tag_index();
      $insert_size_objs = [ grep { defined $_->tag_index() && ( $_->tag_index() == $tag_index ) } @{$insert_size_objs} ];
   }else{
      $insert_size_objs = [ grep { ! defined $_->tag_index() } @{$insert_size_objs} ];
   }

   my $num_insert_objs = scalar @{$insert_size_objs};
   if ($num_insert_objs == 0 )  {
          return 0;
   }elsif( $num_insert_objs > 1 ){
          croak 'There are more than one insert_size objects available';
   }

   my $median = $insert_size_objs->[0]->median();

   if(!$median){
      return 0;
   }
   return $median;
}

=head2 process

process

=cut
sub process { ## no critic (Subroutines::ProhibitExcessComplexit)
  my $self = shift;

  my $id_run = $self->id_run();
  $self->log("Generating bam header for run $id_run");

  my @program_list =();

  my $rta = $self->schema_info->{SCS};
  if(! (scalar keys %{$rta}) ){
      $rta = {ID => 'SCS', DS => 'Controlling software on instrument',};
  }
  push @program_list, $rta;

  my $basecalling = $self->schema_info->{basecalling};
  if(! (scalar keys %{$basecalling}) ){
      $basecalling = {ID => 'basecalling',};
  }
  push @program_list, $basecalling;

  my $bcl_to_qseq = $self->schema_info->{bcl_to_qseq};
  if( scalar keys %{ $bcl_to_qseq } ){
    push @program_list, $bcl_to_qseq;
  }

  my $demultiplex = $self->schema_info->{demultiplex};
  if( scalar keys %{ $demultiplex } ){
    push @program_list, $demultiplex;
  }

  my $recalibration_alignment = $self->schema_info->{recalibration_alignment};
  if( scalar keys %{ $recalibration_alignment } && $recalibration_alignment->{CL} ne q{not run} ){
    push @program_list, $recalibration_alignment;
  }

  my $recalibration_table_creation = $self->schema_info->{recalibration_table_creation};
  if( scalar keys %{ $recalibration_table_creation } && $recalibration_table_creation->{CL} ne q{not run} ){
    push @program_list, $recalibration_table_creation;
  }

  my $recalibration = $self->recalibration();
  if( scalar keys %{$recalibration} && $recalibration->{CL} ne q{not run} ){
    push @program_list, $self->recalibration();
  }

  my $fastq_creator = $self->schema_info->{fastq_creator};
  if( scalar keys %{ $fastq_creator } ){
    push @program_list, $fastq_creator;
  }

  my $split_phix = $self->schema_info->{split_spiked_phix};
  if( scalar keys %{ $split_phix } ){
    push @program_list, $split_phix;
  }

  my $split_nonconsented_sequence = $self->schema_info->{split_nonconsented_sequence};
  if( scalar keys %{ $split_nonconsented_sequence } ){
    push @program_list, $split_nonconsented_sequence;
  }

  if(defined $self->tag_index()){
     push @program_list, $self->split_fastq_by_tag();
  }

  if(scalar keys %{$self->bwa_aln()}){

    push @program_list, $self->bwa_aln();
    push @program_list, $self->bwa_sam();
    push @program_list, $self->sam_fastq_check();
  }elsif($self->fastq2sam()){

     push @program_list, $self->fastq2sam();
  }

  if(scalar keys %{$self->sam_original_quality()}){
     push @program_list, $self->sam_original_quality();
  }

  if(scalar keys %{$self->sam_second_call()}){
     push @program_list, $self->sam_second_call();
  }

  if(scalar keys %{$self->sam_index_tag()}){
    push @program_list, $self->sam_index_tag();
  }
  push @program_list, $self->sam_header();

  if($self->picard_converter_cmd){
    push @program_list, $self->picard();
  }

  if($self->samtools_cmd){
    push @program_list, $self->samtools_fixmate();
    push @program_list, $self->samtools_sort();
  }

  push @program_list, $self->picard_markduplicates();

  $self->_print_pg(\@program_list);

  my $rg_id = 1;
  my $rg_header = $self->_construct_rg($rg_id);
  print $rg_header, "\n" or croak 'cannot print on stdout';

  my $num_sq_lines;
  if( ! $self->no_alignment() ) {
    #pring @SQ header from dictionary now
    $num_sq_lines =   $self->_parsing_picard_dict();
  }

  my $sam_file = $self->sam_file();
  my $sam_fh;

  #sam from a file or stdin
  if($sam_file){
     open $sam_fh, q{<}, $sam_file or croak "Can not open file: $sam_file"; ## no critic (InputOutput::RequireBriefOpen)
  }else{
     $sam_fh =q{STDIN};
  }
  my $line;
  while($line = <$sam_fh>){

    if($line !~ /^@/mxs){
      chomp $line;
      print "$line\tRG:Z:$rg_id\n" or croak 'cannot print on stdout';
    }elsif ( ($line =~ /^\@SQ/mxs) && $num_sq_lines ){
      next;
    }else{
      print "$line" or croak 'cannot print on stdout';
    }
  }
  close $sam_fh or croak "Can not close file: $sam_file";

  return 1;
}

sub _print_pg {
  my ($self, $program_list) = @_;
  my $pp;
  foreach my $program (@{$program_list}){

     my $id = $program->{ID};
     my $pn = $program->{PN};
     my $vn = $program->{VN};
     my $cl = $program->{CL};
     my $ds = $program->{DS};

     print q{@}.qq{PG\tID:$id} or croak 'cannot print on stdout';

     if($pn){
       print qq{\tPN:$pn} or croak 'cannot print on stdout';
     }

     if($vn){
       $vn = $self->_check_tag_value( $vn );
       print qq{\tVN:$vn} or croak 'cannot print on stdout';
     }

     if($cl){
       $cl = $self->_check_tag_value( $cl );
       print qq{\tCL:$cl} or croak 'cannot print on stdout';
     }

     if($ds){
       print qq{\tDS:$ds} or croak 'cannot print on stdout';
     }

     if($pp){
       print qq{\tPP:$pp} or croak 'cannot print on stdout';
     }

     print qq{\n} or croak 'cannot print on stdout';
     $pp = $id;
  }
  return 1;
}

sub _construct_rg {
   my ($self, $id) = @_;

   my $rg_header = q{@}.qq{RG\tID:$id};
   $rg_header   .= qq{\tPL:$SEQUENCING_PLATFORM};
   $rg_header   .= qq{\tCN:$SEQUENCING_CENTER};

   my $library;
   eval{
      $library = $self->asset();
      1;
   } or do {
      carp $EVAL_ERROR;
   };

   my $sample=[];
   eval{
     $sample = $self->samples();
     1;
   } or do {
      carp $EVAL_ERROR;
   };

   my @studies;
   eval{
      my $entity = $self->st_lane()->entity();
      my $study = $entity->study();
      if($study){
         push @studies, $study;
      }
   } or do{
      carp $EVAL_ERROR;
   };

   my $study_name;

   if(scalar @studies){
      $study_name = join q{. },  map { q{Study }.$_->publishable_name. q{: } . $_->get(q{Study description})->[0] } @studies;
   }

   my $library_name;
   if($library){
     $library_name = $library->id();
   }

   my $sample_name = q{unknown};
   if($sample && (scalar @{$sample} == 1)){
     $sample_name = $sample->[0]->publishable_name();
   }elsif($library_name){
     $sample_name = $library_name;
   }
   $sample_name = $self->_check_tag_value( $sample_name );
   $rg_header .= qq{\tSM:}.$sample_name;

   if($library_name){
     $library_name = $self->_check_tag_value( $library_name );
     $rg_header .= qq{\tLB:}. $library_name;
   }

   my $median_insert_size = $self->median_insert_size();
   if( $median_insert_size ) {
     $rg_header .= qq{\tPI:}.$median_insert_size;
   }

   my $run_create_date;
   my $run_folder;
   eval{
    my $api_run = npg::api::run->new({id_run => $self->id_run()});
    $run_create_date = $api_run->loading_date();
    $run_folder = $api_run->run_folder();
    1;
   } or do {
      croak $EVAL_ERROR;
   };

   if ($run_create_date){
     $rg_header .= qq{\tDT:}.$run_create_date;
   }

   if(!$run_folder){
      $run_folder = $self->id_run();
   }

   my $pu = $run_folder.q{_}.$self->position();
   if($self->human_split){
     $pu .= q{_}.$self->human_split;
   }
   if(defined $self->tag_index ){
     $pu .= q{#}.$self->tag_index;
   }
   $rg_header .= qq{\tPU:}.$pu;

   if( $study_name ){
     $study_name = $self->_check_tag_value( $study_name );
     $rg_header .= qq{\tDS:}. $study_name;
   }

   return $rg_header;
}

sub _check_tag_value {
  my ($self, $tag_value) = @_;

  $tag_value =~ s/\n/\ /gmxs;
  $tag_value =~ s/\t/\ /gmxs;

  return $tag_value;
}

sub _parsing_picard_dict {
  my $self = shift;

  my $picard_dict_file = $self->picard_dict();

  if( (!$picard_dict_file ) || (! -e $picard_dict_file ) ){
    return 0;
  }

  my $num_sq_lines;

  open my $picard_dict_fh, q{<}, $picard_dict_file or croak "Can not open file: $picard_dict_file";

  while (my $line = <$picard_dict_fh> ){
    if( $line =~ /^\@SQ/mxs ) {
       print $line or croak 'cannot print on stdout';
       $num_sq_lines ++;
    }
  }

  close $picard_dict_fh or croak "Can not close file: $picard_dict_file";

  return $num_sq_lines;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=item File::Basename

=item File::Spec

=item Readonly

=item XML::LibXML

=item npg_common::roles::run::short_info

=item npg_common::roles::run::path_info

=item npg_common::roles::log

=item npg_common::roles::run::long_info

=item npg_qc::autoqc::checks::check

=item npg_common::sequence::SAM_Second_Call

=item npg_common::sequence::Fastq2SAM

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
