use strict;
use warnings;
use Carp;
use English qw{-no_match_vars};
use Test::More 'no_plan';#tests => 5;
use lib qw{t};

SKIP:
{
  eval {
    croak if !$ENV{MQ_TESTS};
  };
  skip (q{MQ_TESTS not set}, 1) if $EVAL_ERROR;
  
use_ok('npg_common::mq::stomp::queue');

## Durable subscriptions to a queue
{
  my $queue = npg_common::mq::stomp::queue->new();
  my $subscription;
  eval { $subscription = $queue->subscribe({
    queue => q{/queue/songs/by/queen},
    client_id => q{freddie_mercury},
    ack => q{client},
    retries => 2,
  }); };
  is($EVAL_ERROR, q{}, q{no croak subscribing with a durable subscriber to selected queue});
  $queue->delete_connection();
}
{
  my $queue = npg_common::mq::stomp::queue->new();
  my $subscription;
  eval { $subscription = $queue->subscribe({
    destination => q{/queue/songs/by/queen},
    client_id => q{brian_may},
    ack => q{client},
  }); };
  is($EVAL_ERROR, q{}, q{no croak subscribing with a durable subscriber to selected queue});
  $queue->delete_connection();
}
{
  my $queue = npg_common::mq::stomp::queue->new();
  my $subscription;
  eval { $subscription = $queue->subscribe(); };
  is($EVAL_ERROR, q{}, q{no croak subscribing to defaults});
  $queue->delete_connection();
}
## sending a message to a queue
{
  my $queue = npg_common::mq::stomp::queue->new();
  isa_ok($queue, q{npg_common::mq::stomp::queue}, q{$queue});
  eval { $queue->send({}); };
  like($EVAL_ERROR, qr{No[ ]message[ ]text}, q{croak as no message/body text});
}

{
  my $queue = npg_common::mq::stomp::queue->new();

  my $arg_refs = {
    queue   => q{/queue/songs/by/queen},
    body => q{Bohemian Rhapsody},
  };
  eval { $queue->send($arg_refs); };
  is($EVAL_ERROR, q{}, q{no croak sending - queue used for destination});
}

{
  my $queue = npg_common::mq::stomp::queue->new();

  my $arg_refs = {
    destination => q{/queue/songs/by/queen},
    message => q{Killer Queen},
  };
  eval { $queue->send($arg_refs); };
  is($EVAL_ERROR, q{}, q{no croak sending - destination used for destination});
}

## getting messages from a queue using durable subscribers
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    client_id => q{brian_may},
    ack => q{client},
  }); };
  is($EVAL_ERROR, q{}, q{no croak getting message with a durable subscriber to selected queue});
  is($message, q{Bohemian Rhapsody}, q{Bohemian Rhapsody first message given to brian_may});
}

{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    client_id => q{freddie_mercury},
    ack => q{client},
  }); };
  is($EVAL_ERROR, q{}, q{no croak subscribing with a durable subscriber to selected queue});
  is($message, q{Killer Queen}, q{Killer Queen first message given to freddie_mercury});
}

{
  my $queue = npg_common::mq::stomp::queue->new();

  my $arg_refs = {
    queue   => q{/queue/songs/by/queen},
    body => q{A Kind of Magic},
  };
  eval { $queue->send($arg_refs); };
  is($EVAL_ERROR, q{}, q{no croak sending - queue used for destination});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    client_id => q{freddie_mercury},
    ack => q{client},
  }); };
  is($EVAL_ERROR, q{}, q{no croak subscribing with a durable subscriber to selected queue});
  is($message, q{A Kind of Magic}, q{A Kind of Magic second message given to freddie_mercury});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    durable => q{true},
    client_id => q{brian_may},
    ack => q{client},
  }); };
  is($EVAL_ERROR, q{}, q{no croak getting message with a durable subscriber to selected queue});
  is($message, undef, q{No message on queue to be given to brian_may});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $arg_refs = {
    queue   => q{/queue/songs/by/queen},
    body => q{One Vision},
  };
  eval { $queue->send($arg_refs); };
  is($EVAL_ERROR, q{}, q{no croak sending - queue used for destination});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    durable => q{true},
    client_id => q{freddie_mercury},
    ack => q{client},
    code => sub { my $message = shift; print "$message\n"; return { message => $message }; },
    retries => 2,
  }); };
  is($EVAL_ERROR, q{}, q{no croak getting message with a durable subscriber to selected queue});
  is(ref$message, q{HASH}, q{One Vision third message processed by code block and return value of block obtained});
  is($message->{message}, q{One Vision}, q{message in returned data structure ok});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $arg_refs = {
    queue   => q{/queue/songs/by/queen},
    body => q{Hammer to Fall},
  };
  eval { $queue->send($arg_refs); };
  is($EVAL_ERROR, q{}, q{no croak sending - queue used for destination});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    durable => q{true},
    client_id => q{freddie_mercury},
    ack => q{client},
    code => sub { my $message = shift; croak q{Deliberately croaking, to see what happens.}; },
    retries => 2,
  }); };
  is($EVAL_ERROR, q{}, q{no croak getting message with code block which deliberately croaks as freddie_mercury});
  is($message, undef, q{no message or output obtained as code croaked});
}
{
  my $queue = npg_common::mq::stomp::queue->new();

  my $message;
  eval { $message = $queue->get_message({
    destination => q{/queue/songs/by/queen},
    client_id => q{freddie_mercury},
    ack => q{client},
  }); };
  is($EVAL_ERROR, q{}, q{no croak getting message with freddie_mercury to selected queue});
  is($message, q{Hammer to Fall}, q{message Hammer to Fall obtained off queue, so was left there});
}

} # end of total SKIP block
1;
